// vim: set tabstop=4 shiftwidth=4 expandtab
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sudoku.h"

enum cell_bitset {
    CELL_BITSET_FULL = 0x1ff,
    CELL_BITSET_INVALID = 0xffff,
};

/* ************************************************************** *
 *                 Advised auxiliary functionns                   *
 * ************************************************************** */

static unsigned int bitset_add(unsigned int original, int number)
{
    assert(number >= 1 && number <= 9);
    return original | (1u << (number - 1));
}

static bool bitset_is_set(unsigned int original, int query)
{
    assert(query >= 1 && query <= 9);
    return original & (1u << (query - 1));
}

static bool bitset_is_unique(unsigned int original)
{
    return original != 0 && (original & (original - 1)) == 0;
}

static int bitset_next(unsigned int bitset, int previous)
{
    assert(previous <= 9);
    if (previous <= 0)
        previous = 0;

    for (int i = previous; i < 9; ++i)
        if (bitset_is_set(bitset, i + 1))
            return i + 1;

    return -1;
}

static unsigned int compute_bitset(unsigned int sudoku[9][9], int min_row, int max_row, int min_col, int max_col, bool report_dup)
{
    assert(min_row >= 0 && min_row < 9);
    assert(max_row >= 0 && max_row < 9);
    assert(min_col >= 0 && min_col < 9);
    assert(max_col >= 0 && max_col < 9);

    unsigned int masked = 0u;

    for (int row = min_row; row <= max_row; ++row) {
        for (int col = min_col; col <= max_col; ++col) {
            if (!bitset_is_unique(sudoku[row][col]))
                continue;

            if (report_dup && (masked & sudoku[row][col]))
                return CELL_BITSET_INVALID;

            masked |= sudoku[row][col];
        }
    }

    assert(masked <= CELL_BITSET_FULL);
    return masked;
}

static unsigned int eliminate_any(unsigned int sudoku[9][9], int min_row, int max_row, int min_col, int max_col)
{
    assert(min_row >= 0 && min_row < 9);
    assert(max_row >= 0 && max_row < 9);
    assert(min_col >= 0 && min_col < 9);
    assert(max_col >= 0 && max_col < 9);

    unsigned int masked = compute_bitset(sudoku, min_row, max_row, min_col, max_col, false);
    bool changed = false;

    for (int row = min_row; row <= max_row; ++row) {
        for (int col = min_col; col <= max_col; ++col) {
            if (bitset_is_unique(sudoku[row][col]))
                continue;

            changed |= masked & sudoku[row][col];
            sudoku[row][col] &= ~masked;
        }
    }

    return changed;
}

bool eliminate_row(unsigned int sudoku[9][9], int row_index)
{
    assert(row_index >= 0 && row_index < 9);
    return eliminate_any(sudoku, row_index, row_index, 0, 8);
}

bool eliminate_col(unsigned int sudoku[9][9], int col_index)
{
    assert(col_index >= 0 && col_index < 9);
    return eliminate_any(sudoku, 0, 8, col_index, col_index);
}

bool eliminate_box(unsigned int sudoku[9][9], int row_index, int col_index)
{
    assert(row_index >= 0 && row_index < 9);
    assert(col_index >= 0 && col_index < 9);
    // set indices to start of the current box
    row_index -= row_index % 3;
    col_index -= col_index % 3;
    return eliminate_any(sudoku, row_index, row_index + 2, col_index, col_index + 2);
}

bool needs_solving(unsigned int sudoku[9][9])
{
    for (int row = 0; row < 9; ++row) {
        for (int col = 0; col < 9; ++col) {
            if (!bitset_is_unique(sudoku[row][col]))
                return true;
        }
    }
    return false;
}

bool is_valid(unsigned int sudoku[9][9])
{
    for (int row = 0; row < 3; ++row) {
        for (int col = 0; col < 3; ++col) {
            unsigned int box = compute_bitset(sudoku, 3 * row, 3 * row + 2, 3 * col, 3 * col + 2, true);
            if (box == CELL_BITSET_INVALID)
                return false;
        }
    }

    for (int index = 0; index < 9; ++index) {
        unsigned int row_bitset = compute_bitset(sudoku, index, index, 0, 8, true);
        if (row_bitset == CELL_BITSET_INVALID)
            return false;

        unsigned int col_bitset = compute_bitset(sudoku, 0, 8, index, index, true);
        if (col_bitset == CELL_BITSET_INVALID)
            return false;
    }

    for (int row = 0; row < 9; ++row) {
        for (int col = 0; col < 9; ++col) {
            unsigned int cell = sudoku[row][col];
            if (bitset_is_unique(cell))
                continue;

            if (cell == 0)
                return false;
        }
    }

    return true;
}

bool solve(unsigned int sudoku[9][9])
{
    bool valid = false;
    bool changed = true;

    if (!needs_solving(sudoku))
        return is_valid(sudoku);

    while ((valid = is_valid(sudoku)) && changed) {
        changed = false;
        for (int row = 0; row < 9; ++row) {
            changed |= eliminate_row(sudoku, row);
        }

        for (int col = 0; col < 9; ++col) {
            changed |= eliminate_col(sudoku, col);
        }

        for (int row = 0; row < 3; ++row) {
            for (int col = 0; col < 3; ++col) {
                changed |= eliminate_box(sudoku, 3 * row, 3 * col);
            }
        }
    }

    return !needs_solving(sudoku) && valid;
}

static bool store(unsigned int sudoku[9][9], int row, int col, int c, bool allow_dot)
{
    if ((c == '0') || (allow_dot && c == '.')) {
        sudoku[row][col] = CELL_BITSET_FULL;
        return true;
    }

    if (c > '0' && c <= '9') {
        sudoku[row][col] = bitset_add(0u, c - '0');
        return true;
    }

    return false;
}

static bool load_raw(unsigned int sudoku[9][9], int c)
{
    for (int row = 0; row < 9; ++row) {
        for (int col = 0; col < 9; ++col) {
            if (!store(sudoku, row, col, c, false)) {
                fprintf(stderr, "Failed to load input\n");
                return false;
            }

            c = getchar();
        }
    }

    bool loaded_ok = c == '\n' || c != EOF;
    if (!loaded_ok)
        fprintf(stderr, "Failed to load input - garbage at the end of sudoku\n");
    return loaded_ok;
}

static bool load_graphical_row_core(unsigned int sudoku[9][9], int c, int row)
{
    if (c != '|')
        return false;

    for (int group = 0; group < 3; ++group) {
        c = getchar();
        if (c != ' ')
            return false;

        for (int col = 3 * group; col < 3 * (group + 1); ++col) {
            c = getchar();
            if (!store(sudoku, row, col, c, true))
                return false;

            c = getchar();
            if (c != ' ')
                return false;
        }

        c = getchar();
        if (c != '|')
            return false;
    }

    c = getchar();
    return c == '\n';
}

static bool load_graphical_row(unsigned int sudoku[9][9], int c, int row)
{
    bool rv = load_graphical_row_core(sudoku, c, row);
    if (!rv)
        fprintf(stderr, "Failed to load row\n");
    return rv;
}

static bool load_graphical_line(int c)
{
    // todo: implementation through comparing individual characters
    // to expected input "+=======+=======+=======+" would no doubt be
    // a better move. Yet provide different approach.

    if (c != '+')
        return false;

    for (int group = 0; group < 3; ++group) {
        for (int col = 3 * group; col < 3 * (group + 1); ++col) {
            for (int subcol = 0; subcol < 2; ++subcol) {
                c = getchar();
                if (c != '=') {
                    return false;
                }
            }
        }
        c = getchar();
        if (c != '=') {
            return false;
        }

        c = getchar();
        if (c != '+')
            return false;
    }

    c = getchar();
    return c == '\n';
}

static bool load_graphical(unsigned int sudoku[9][9], int c)
{
    if (!load_graphical_line(c)) {
        fprintf(stderr, "Failed to load input - incorrect format at the start of the input\n");
        return false;
    }

    for (int row_group = 0; row_group < 3; ++row_group) {
        for (int row = 3 * row_group; row < 3 * (row_group + 1); ++row) {
            if (!load_graphical_row(sudoku, getchar(), row)) {
                return false;
            }
        }
        if (!load_graphical_line(getchar())) {
            fprintf(stderr, "Failed to load input - incorrect box separator\n");
            return false;
        }
    }

    return true;
}

bool load(unsigned int sudoku[9][9])
{
    int c = '\n';
    do {
        c = getchar();
    } while (c == '\n');

    if (c == '+') {
        return load_graphical(sudoku, c);
    }

    if ('0' <= c && c <= '9') {
        return load_raw(sudoku, c);
    }

    fprintf(stderr, "Unrecognized input format\n");
    return false;
}

static void print_graphical_cell(unsigned int cell)
{
    if (cell == 0) {
        putchar('!');
        return;
    }

    if (!bitset_is_unique(cell)) {
        putchar('.');
        return;
    }

    int number = bitset_next(cell, 0);
    printf("%d", number);
}

static void print_graphical_row(unsigned int sudoku[9][9], int row)
{
    putchar('|');

    for (int group = 0; group < 3; ++group) {
        for (int col = 3 * group; col < 3 * (group + 1); ++col) {
            putchar(' ');
            print_graphical_cell(sudoku[row][col]);
        }
        printf(" |");
    }
    putchar('\n');
}

static void print_graphical_line(void)
{
    puts("+=======+=======+=======+");
}

void print(unsigned int sudoku[9][9])
{
    print_graphical_line();

    for (int row_group = 0; row_group < 3; ++row_group) {
        for (int row = 3 * row_group; row < 3 * (row_group + 1); ++row)
            print_graphical_row(sudoku, row);
        print_graphical_line();
    }
}

/* ************************************************************** *
 *                              Bonus                             *
 * ************************************************************** */

#ifdef BONUS_GENERATE
static int real_pos(int pos, bool available_fields[])
{
    for (int rpos = pos; rpos < 81; ++rpos) {
        if (available_fields[rpos])
            return rpos;
    }
    for (int lpos = pos - 1; lpos >= 0; --lpos) {
        if (available_fields[lpos])
            return lpos;
    }
    return -1;
}

void generate(unsigned int sudoku[9][9])
{
    unsigned int sudoku_dirty[9][9];
    memcpy(sudoku_dirty, sudoku, sizeof(sudoku_dirty));

    if (!solve(sudoku_dirty))
        return;

    bool available_fields[9 * 9] = { false };
    for (int x = 0; x < 81; ++x) {
        int row = x / 9;
        int col = x % 9;

        available_fields[x] = bitset_is_unique(sudoku[row][col]);
        if (!available_fields[x])
            sudoku[row][col] = CELL_BITSET_FULL;
    }

    while (true) {
        int pos = real_pos(rand() % 81, available_fields);
        if (pos < 0)
            return; // no more options to cut

        available_fields[pos] = false;
        memcpy(sudoku_dirty, sudoku, sizeof(sudoku_dirty));

        int row = pos / 9;
        int col = pos % 9;
        sudoku_dirty[row][col] = CELL_BITSET_FULL;

        if (!solve(sudoku_dirty))
            continue;

        sudoku[row][col] = CELL_BITSET_FULL;
    }

    return;
}
#endif

#ifdef BONUS_GENERIC_SOLVE

static bool generic_solve_ws(unsigned int sudoku_in[9][9], unsigned int sudoku_out[9][9], int srow, int scol)
{
    bool elimination_succ = solve(sudoku_in);
    if (elimination_succ) {
        if (sudoku_in != sudoku_out)
            memcpy(sudoku_out, sudoku_in, 9 * 9 * sizeof(sudoku_in[0][0]));
        return true;
    }

    if (!is_valid(sudoku_in))
        return false;

    for (int row = srow; row < 9; row++) {
        for (int col = scol; col < 9; col++) {
            if (bitset_is_unique(sudoku_in[row][col]))
                continue;

            unsigned int sudoku_tmp[9][9];
            int previous = 0;

            do {
                previous = bitset_next(sudoku_in[row][col], previous);
                if (!(previous > 0))
                    return false;

                memcpy(sudoku_tmp, sudoku_in, 9 * 9 * sizeof(sudoku_in[0][0]));
                sudoku_tmp[row][col] = bitset_add(0, previous);
                eliminate_row(sudoku_tmp, row);
                eliminate_col(sudoku_tmp, col);
                eliminate_box(sudoku_tmp, row, col);

                if (generic_solve_ws(sudoku_tmp, sudoku_out, row, col))
                    return true;

            } while (previous >= 0);
        }
        scol = 0;
    }

    return false;
}

bool generic_solve(unsigned int sudoku[9][9])
{
    return generic_solve_ws(sudoku, sudoku, 0, 0);
}
#endif
